@@ -21,12 +21,107 @@ from libopensesame.py3compat import *
 import os
 import sys
 import platform
-os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"
+import multiprocessing
+
+# Appears more stable when running experiments on Linux, where fork is the 
+# default multiprocessing method.
+if platform.system() == 'Linux':
+	multiprocessing.set_start_method('spawn')
+# On Linux this appears to be buggy
+if platform.system() != 'Linux':
+	os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"
+# solves a library conflict for Linux with Nvidia drivers
+# See https://forum.qt.io/topic/81328/ubuntu-qopenglshaderprogram-shader-program-is-not-linked/2
+else:
+	try:
+		from OpenGL import GL
+	except ImportError:  # Rapunzel doesn't need OpenGL
+		pass
+
+# Attach a dummy console when launched with pythonw.exe. This is necessary for
+# some libraries.
+if sys.executable.endswith('pythonw.exe'):
+	sys.stdout = open(os.devnull, 'w')
+	sys.stderr = open(os.devnull, 'w')
+	sys.stdin = open(os.devnull)
+
+# On MacOS, the locations of the SSL cerfificates need to be explicitly indicated when packaged
+# as an app, otherwise urllib.urlopen fails with an SSL related error. To do so, we make use of the
+# certifi package that can provide this location, but it is unclear if certifi is always installed,
+# so let's fail gracefully if it isn't.
+if platform.system() == 'Darwin':
+	try:
+		import certifi
+		os.environ['SSL_CERT_FILE'] = certifi.where()
+		os.environ['SSL_CERT_DIR'] = os.path.dirname(certifi.where())
+	except ImportError:
+		pass # Or log something here?
+	# Workaround for an incompatibility between Qt and Mac OS Big Sur. See:
+	# - https://forum.cogsci.nl/discussion/comment/21525/#Comment_21525
+	os.environ['QT_MAC_WANTS_LAYER'] = '1'
+
+def patch_pyqt():
+
+	"""This patches PyQt such that properties are removed from objects before
+	connectSlotsByName() tries to inspect the object. This is necessary because
+	in some versions, the introspection crashes when properties cannot be
+	accessed, for example because the object needs to be initialized first.
+	"""
+
+	def _(fnc):
+
+		def make_object_safe(obj, cls):
+
+			tmp = {}
+			for name, value in list(cls.__dict__.items()):
+				try:
+					getattr(obj, name)
+				except AttributeError:
+					tmp[name] = value
+					delattr(cls, name)
+			return tmp
+
+		def restore_object(cls, tmp):
+
+			for name, value in tmp.items():
+				setattr(cls, name, value)
+
+		def inner(obj):
+
+			tmp = {}
+			for cls in obj.__class__.mro():
+				tmp[cls] = make_object_safe(obj, cls)
+			fnc(obj)
+			for cls in tmp:
+				restore_object(cls, tmp[cls])
+
+		return inner
+
+	from qtpy import QtCore
+	QtCore.QMetaObject.connectSlotsByName = \
+		_(QtCore.QMetaObject.connectSlotsByName)
+
 
 def set_paths():
 
-	from qtpy import QtCore
 	from libopensesame import misc
+	import sys
+	import platform
+	# On Windows, the working directory should be the folder with the Python
+	# interpreter. If not, Qt will have difficulty finding the correct paths.
+	# We can disable this behavior on the command line, because it will
+	# actually break things on some installations.
+	if platform.system() == 'Windows' and '--no-chdir' not in sys.argv:
+		os.chdir(os.path.dirname(sys.executable))
+	# Update the system path so that Qt can find itself
+	path = os.path.join(os.getcwd(), 'Library', 'bin')
+	if os.path.exists(path):
+		os.environ['PATH'] = path + ';' + os.environ['PATH']
+	# Add the Scripts subfolder, which is where Anaconda scripts are located
+	path = os.path.join(os.getcwd(), 'Scripts')
+	if os.path.exists(path):
+		os.environ['PATH'] = path + ';' + os.environ['PATH']
+	from qtpy import QtCore
 	# Add the folder that contains the OpenSesame modules to the path. This is
 	# generally only necessary if OpenSesame is directly run from source,
 	# instead from an installation.
@@ -52,6 +147,8 @@ def set_paths():
 def opensesame():
 
 	set_paths()
+	if py3:
+		patch_pyqt()
 	# Support for multiprocessing when packaged
 	# In OS X the multiprocessing module is horribly broken, but a fixed
 	# version has been released as the 'billiard' module
@@ -85,23 +182,28 @@ def opensesame():
 		from qtpy import QtWebEngineWidgets
 	except ImportError:
 		pass
-	app = QApplication(sys.argv)
-	# Enable High DPI pixmaps with PyQt5
-	if hasattr(qtpy.QtCore.Qt, 'AA_UseHighDpiPixmaps'):
-		app.setAttribute(qtpy.QtCore.Qt.AA_UseHighDpiPixmaps)
-	# Enable scaling for High DPI displays with PyQt5
-	if hasattr(qtpy.QtCore.Qt, 'AA_EnableHighDpiScaling'):
-		app.setAttribute(qtpy.QtCore.Qt.AA_EnableHighDpiScaling)
+	if platform.system() != 'Linux':
+		# Enable High DPI pixmaps with PyQt5
+		if hasattr(qtpy.QtCore.Qt, 'AA_UseHighDpiPixmaps'):
+			QApplication.setAttribute(qtpy.QtCore.Qt.AA_UseHighDpiPixmaps)
+		# Enable scaling for High DPI displays with PyQt5
+		if hasattr(qtpy.QtCore.Qt, 'AA_EnableHighDpiScaling'):
+			QApplication.setAttribute(qtpy.QtCore.Qt.AA_EnableHighDpiScaling)
 	from libqtopensesame.qtopensesame import qtopensesame
+	app = QApplication(sys.argv)
 	opensesame = qtopensesame(app)
 	opensesame.__script__ = __file__
 	app.processEvents()
-	# Install the translator. For some reason, the translator needs to be
+	# Install the translator. For some reason, the translators need to be
 	# instantiated here and not in the set_locale() function, otherwise the
 	# locale is ignored.
 	from qtpy.QtCore import QTranslator
-	translator = QTranslator()
-	opensesame.set_locale(translator)
+	opensesame.set_locale([
+		QTranslator(),
+		QTranslator(),
+		QTranslator(),
+		QTranslator()
+	])
 	# Now that the window is shown, load the remaining modules and resume the
 	# GUI initialization.
 	opensesame.resume_init()
@@ -156,23 +258,31 @@ def opensesamerun():
 			sys.exit()
 	# Decode the experiment path and logfile
 	experiment = os.path.abspath(options.experiment)
-	if isinstance(experiment, str):
-		experiment = safe_decode(experiment,
-			enc=libopensesame.misc.filesystem_encoding(), errors=u'ignore')
+	if isinstance(experiment, bytes):
+		experiment = safe_decode(
+			experiment,
+			enc=libopensesame.misc.filesystem_encoding(),
+			errors=u'ignore'
+		)
 	# experiment_path = os.path.dirname(experiment)
 	logfile = options.logfile
-	if isinstance(logfile, str):
-		logfile = safe_decode(logfile,
-			enc=libopensesame.misc.filesystem_encoding(), errors=u'ignore')
-
+	if isinstance(logfile, bytes):
+		logfile = safe_decode(
+			logfile,
+			enc=libopensesame.misc.filesystem_encoding(),
+			errors=u'ignore'
+		)
 	experiment_path = safe_decode(
 		os.path.abspath(options.experiment),
 		enc=libopensesame.misc.filesystem_encoding()
 	)
 	if options.debug:
 		# In debug mode, don't try to catch any exceptions
-		exp = libopensesame.experiment.experiment(u"Experiment",
-			experiment, experiment_path=experiment_path)
+		exp = libopensesame.experiment.experiment(
+			u"Experiment",
+			experiment,
+			experiment_path=experiment_path
+		)
 		exp.set_subject(options.subject)
 		exp.var.fullscreen = options.fullscreen
 		exp.logfile = logfile
@@ -181,11 +291,16 @@ def opensesamerun():
 	else:
 		# Try to parse the experiment from a file
 		try:
-			exp = libopensesame.experiment.experiment(u"Experiment",
-				experiment, experiment_path=experiment_path)
+			exp = libopensesame.experiment.experiment(
+				u"Experiment",
+				experiment,
+				experiment_path=experiment_path
+			)
 		except Exception as e:
-			libopensesame.misc.messagebox(u"OpenSesame Run",
-				libopensesame.misc.strip_tags(e))
+			libopensesame.misc.messagebox(
+				u"OpenSesame Run",
+				libopensesame.misc.strip_tags(e)
+			)
 			sys.exit()
 		# Set some options
 		exp.set_subject(options.subject)
@@ -203,8 +318,12 @@ def opensesamerun():
 			try:
 				exp.end()
 			except Exception as f:
-				libopensesame.misc.messagebox(u"OpenSesame Run",
-					libopensesame.misc.strip_tags(f))
-			libopensesame.misc.messagebox(u"OpenSesame Run",
-				libopensesame.misc.strip_tags(e))
+				libopensesame.misc.messagebox(
+					u"OpenSesame Run",
+					libopensesame.misc.strip_tags(f)
+				)
+			libopensesame.misc.messagebox(
+				u"OpenSesame Run",
+				libopensesame.misc.strip_tags(e)
+			)
 	libopensesame.experiment.clean_up(exp.debug)
